# -*- coding: utf-8 -*-
# YDbf - Yielded dbf reader-writer
# Inspired by code of Raymond Hettinger
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/362715
#
# Copyright (C) 2006 Yury Yurevich, Alexandr Zamaraev
# Copyright (C) 2007-2008 Yury Yurevich
#
# http://www.pyobject.ru/projects/ydbf/
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
Unit-tests for YDbf
"""

import datetime
import unittest
from cStringIO import StringIO

from ydbf import YDbfReader, YDbfWriter
from ydbf.lib import date2dbf, str2dbf, dbf2date, dbf2str

class TestDateConverters(unittest.TestCase):
        
    def test_dbf2date(self):
        self.assertEqual(dbf2date(''), None)
        self.assertEqual(dbf2date('None'), None)
        self.assertEqual(dbf2date(None), None)
        self.assertEqual(dbf2date('20060506'), datetime.date(2006, 5, 6))
        self.assertEqual(dbf2date('0'), None)
        self.assertEqual(dbf2date('000'), None)
        
    def test_date2dbf(self):
        self.assertEqual(date2dbf(datetime.date(2006, 5, 6)), '20060506')
        self.assertRaises(TypeError, date2dbf, None)
        self.assertRaises(TypeError, date2dbf, '20060506')
        
    def test_dbf2str(self):
        self.assertEqual(dbf2str(''), None)
        self.assertEqual(dbf2str('None'), None)
        self.assertEqual(dbf2str(None), None)
        self.assertEqual(dbf2str('20060506'), '06.05.2006')
        self.assertEqual(dbf2str('0'), None)
        self.assertEqual(dbf2str('000'), None)
        
    def test_str2dbf(self):
        self.assertEqual(str2dbf('06.05.2006'), '20060506')
        self.assertRaises(TypeError, str2dbf, None)
        self.assertRaises(TypeError, str2dbf, datetime.date(2200, 1, 1))
        self.assertRaises(ValueError, str2dbf, '')
        self.assertRaises(ValueError, str2dbf, '06/05/2006')


class TestYdbfReader(unittest.TestCase):

    def setUp(self):
        self.dbf_data = '\x03\x06\x06\x13\x03\x00\x00\x00\xc1\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00INT_FLD\x00312N\x05\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00FLT_FLD\x00\x00\x00\x00N\n\x00\x00\x00\x05\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00CHR_FLD\x00\x00\x00\x00C\x10\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00DTE_FLD\x00\x00\x00\x00D\x18\x00\x00\x00\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00BLN_FLD\x00\x00\x00\x00L\x19\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r   2512.34test  20060507T  113 1.01del   20061223F*7436 0.50ex.   20060715T\x1a'

        self.reference_data = [[25, 12.34, 'test',  datetime.date(2006,  5,  7),  True],
                               [113, 1.01,  'del',  datetime.date(2006, 12, 23), False],
                               # skipped deleted line
                              ]
        self.fh = StringIO(self.dbf_data)
        self.dbf = YDbfReader(self.fh)

    def test_dbf2date(self):
        self.assertEqual(self.dbf.dbf2date, dbf2date)

    def test_header(self):
        self.assertEqual(self.dbf._fields, [('DeletionFlag', 'C', 1, 0),
                                        ('INT_FLD',      'N', 4, 0),
                                        ('FLT_FLD',      'N', 5, 2),
                                        ('CHR_FLD',      'C', 6, 0),
                                        ('DTE_FLD',      'D', 8, 0),
                                        ('BLN_FLD',      'L', 1, 0)])
        self.assertEqual(self.dbf.fields, [('INT_FLD',      'N', 4, 0),
                                       ('FLT_FLD',      'N', 5, 2),
                                       ('CHR_FLD',      'C', 6, 0),
                                       ('DTE_FLD',      'D', 8, 0),
                                       ('BLN_FLD',      'L', 1, 0)])
        self.assertEqual(self.dbf.numrec, 3)
        self.assertEqual(self.dbf.stop_at, 3)
        self.assertEqual(self.dbf.lenheader, 193)
        self.assertEqual(self.dbf.numfields, 5)
        self.assertEqual(self.dbf.recsize, 25)
        self.assertEqual(self.dbf.recfmt, '1s4s5s6s8s1s')
        self.assertEqual(self.dbf.field_names, ['INT_FLD', 'FLT_FLD',
                                            'CHR_FLD', 'DTE_FLD',
                                            'BLN_FLD'])

    def test_len(self):
        self.assertEqual(len(self.dbf), 3)

    def test_call(self):
        self.assertEqual(list(self.dbf()), self.reference_data)
        self.assertEqual(list(self.dbf(start_from=1)),
                         [self.reference_data[1]])
        self.assertEqual(list(self.dbf(start_from=0, limit=1)),
                         [self.reference_data[0]])


class TestYdbfWriter(unittest.TestCase):
    def setUp(self):
        self.dbf_reference_data = '\x03j\x06\x13\x03\x00\x00\x00\xc1\x00\x19\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00INT_FLD\x00\x00\x00\x00N\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00FLT_FLD\x00\x00\x00\x00N\x00\x00\x00\x00\x05\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00CHR_FLD\x00\x00\x00\x00C\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00DTE_FLD\x00\x00\x00\x00D\x00\x00\x00\x00\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00BLN_FLD\x00\x00\x00\x00L\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r   2512.34test  20060507T  113 1.01del   20061223F 7436 0.50ex.   20060715T\x1a'
        self.reference_data = [[25, 12.34, 'test', datetime.date(2006,  5,  7),  True],
                               [113, 1.01,  'del', datetime.date(2006, 12, 23), False],
                               [7436, 0.5,  'ex.', datetime.date(2006,  7, 15),  True],
                              ]
        self.reference_data_dict = [
        {'INT_FLD':  25, 'FLT_FLD':12.34, 'CHR_FLD':'test', 'DTE_FLD':datetime.date(2006,  5,  7), 'BLN_FLD':True},
        {'INT_FLD': 113, 'FLT_FLD': 1.01, 'CHR_FLD': 'del', 'DTE_FLD':datetime.date(2006, 12, 23), 'BLN_FLD':False},
        {'INT_FLD':7436, 'FLT_FLD': 0.5,  'CHR_FLD': 'ex.', 'DTE_FLD':datetime.date(2006,  7, 15), 'BLN_FLD':True},
        ]
        self.fields = [('INT_FLD',      'N', 4, 0),
                       ('FLT_FLD',      'N', 5, 2),
                       ('CHR_FLD',      'C', 6, 0),
                       ('DTE_FLD',      'D', 8, 0),
                       ('BLN_FLD',      'L', 1, 0)]
        self.fh = StringIO()
        self.dbf = YDbfWriter(self.fh, self.fields)
        self.dbf_dict = YDbfWriter(self.fh, self.fields, as_dict=True)
        
    
    def test_header(self):
        self.assertEqual(self.dbf.now, datetime.date.today())
        self.assertEqual(self.dbf.numrec, 0)
        self.assertEqual(self.dbf.lenheader, 193)
        self.assertEqual(self.dbf.recsize, 25)
        self.assertEqual(self.dbf.numfields, 5)
        self.assertEqual(self.dbf.sig, 3)

    def test_date2dbf(self):
        self.assertEqual(self.dbf.date2dbf, date2dbf)
    
    def test_call(self):
        self.dbf.now = datetime.date(2006, 6, 19)
        self.assertEqual(self.dbf.now, datetime.date(2006, 6, 19))
        self.dbf(self.reference_data)
        self.assertEqual(self.dbf.numrec, 3)
        self.fh.seek(0)
        data = self.fh.read()
        self.assertEqual(data, self.dbf_reference_data)
    
    def test_call_as_dict(self):
        self.dbf_dict.now = datetime.date(2006, 6, 19)
        self.assertEqual(self.dbf_dict.now, datetime.date(2006, 6, 19))
        self.dbf_dict(self.reference_data_dict)
        self.assertEqual(self.dbf_dict.numrec, 3)
        self.fh.seek(0)
        data = self.fh.read()
        self.assertEqual(data, self.dbf_reference_data)

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# YDbf - Yielded dbf reader-writer
# Inspired by code of Raymond Hettinger
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/362715
#
# Copyright (C) 2006 Yury Yurevich, Alexandr Zamaraev
# Copyright (C) 2007-2008 Yury Yurevich
#
# http://www.pyobject.ru/projects/ydbf/
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
DBF reader
"""
__all__ = ["YDbfBasicReader", "YDbfStrictReader", "YDbfReader"]

import datetime
import struct
import itertools

from ydbf import lib

class YDbfBasicReader(object):
    """
    Basic class for reading DBF
    
    Instance is an iterator over DBF records
    
    @param fh: filehandler (should be opened for binary reads)
    """
    def __init__(self, fh):
        """
        Constructor
        """
        self.fh = fh             # filehandler
        self.numrec = 0          # number of records
        self.lenheader = 0       # length of header
        self.numfields = 0       # number of fields
        self.fields = []         # fields info in format [(NAME, TYP, SIZE, DEC),]
        self.field_names = ()    # field names (i.e. (NAME,))
        self.start_from = 0      # number of rec, iteration started from
        self.stop_at = 0         # number of rec, iteration stopped at (not include this)
        self.recfmt = ''         # struct-format of rec
        self.recsize = 0         # size of each record (in bytes)
        self.i = 0               # current item in iterator
        self.dt = None           # date of file creation
        self.dbf2date = lib.dbf2date # function for conversion from dbf to date

        self.readHeader()

    def readHeader(self):
        """
        Read DBF header
        """ 
        self.fh.seek(0)

        sig, year, month, day, numrec, lenheader, recsize, lang = struct.unpack(
            lib.HEADER_FORMAT,
            self.fh.read(32))
        year = year + 1900
        # some software use 0x08 as 2008 instead of 0x6c
        if year < 1950:
            year = year+100
        self.dt = datetime.date(year, month, day)
        self.sig = sig
        if sig not in lib.SUPPORTED_SIGNATURES:
            version = lib.SIGNATURES.get(sig, 'UNKNOWN')
            raise lib.DbfError("DBF version '%s' (signature %s) not supported" % (version, hex(sig)))
        
        numfields = (lenheader - 33) // 32
        fields = []
        for fieldno in xrange(numfields):
            name, typ, size, deci = struct.unpack(lib.FIELD_DESCRIPTION_FORMAT,
                                                  self.fh.read(32))
            name = name.split('\0', 1)[0]       # NULL is a end of string
            fields.append((name, typ, size, deci))
        terminator = self.fh.read(1)
        assert terminator == '\x0d', "Terminator must be 0x0d"
                
        fields.insert(0, ('DeletionFlag', 'C', 1, 0))
        self.raw_lang = lang
        self._fields = fields  # with DeletionFlag
        self.fields = fields[1:] # without DeletionFlag
        self.recfmt = ''.join(['%ds' % fld[2] for fld in fields])
        self.recsize = struct.calcsize(self.recfmt)
        self.numrec = numrec
        self.lenheader = lenheader
        self.numfields = numfields
        self.stop_at = numrec
        self.field_names = [fld[0] for fld in self.fields]
    
    def __len__(self):
        """
        Get number of records in DBF
        @return: number of records (integer)
        """
        return self.numrec
    
    def __call__(self, start_from=None, limit=None, raise_on_unknown_type=False):
        """
        Get iterator
        @param start_from: index of record start from (optional)
        @param limit: limits number of iterated records (optional)
        @param raise_on_unknown_type: raise or not exception for unknown type
            of field
        @return: iterator over records
        """
        if start_from is not None:
            self.start_from = start_from
        offset = self.lenheader + self.recsize*self.start_from
        if self.fh.tell() != offset:
            self.fh.seek(offset)
            
        if limit is not None:
            self.stop_at = self.start_from + limit

        logic = {
            'Y': True, 'y': True, 'T': True, 't': True,
            'N': False, 'n': False, 'F': False, 'f': False,
            }
        types_in_list = [t in ('N', 'D', 'L', 'C') for n, t, s, d in self._fields]
        if False in types_in_list and raise_on_unknown_type:
            idx = types_in_list.index(False)
            raise lib.DbfTypeError("Unknown dbf-type: %s in field %s" % \
                    (self._fields[idx][1], self._fields[idx][0]))
        actions = {
            'D': lambda val: self.dbf2date(val.strip()),
            'L': lambda val: logic.get(val.strip()),
            'C': lambda val: val.rstrip(),
            'N': lambda val: (val.strip() or 0) and int(val),
            'ND': lambda val: (val.strip() or 0.0) and float(val),
            'DeletionFlag': None,
            }
        converters = [
                         # first of all, is it DeletionFlag
            actions.get((name == 'DeletionFlag' and 'DeletionFlag') or
                         # secondary, D or C or someting else
                    (typ != 'N' and typ) or
                         # third, decimal (ND)
                    (deci and 'ND') or
                         # the last -- N
                    'N') # default value -- None
                   for name, typ, size, deci in self._fields]
        
        for i in xrange(self.start_from, self.stop_at):
            record = struct.unpack(self.recfmt, self.fh.read(self.recsize))
            self.i = i
            if record[0] != ' ':
                continue                        # deleted record
            try:
                res = [conv(val)
                    for idx, (val, conv) in enumerate(itertools.izip(record, converters))
                    if conv
                    ]
            except (IndexError, ValueError, TypeError), err:
                    raise lib.DbfError("Error occured while reading value '%s' from field '%s' (rec #%d)" % \
                            (val, self._fields[idx][0], self.i))
            yield res

class YDbfStrictReader(YDbfBasicReader):
    """
    DBF-reader with additional checks
    """
    def __init__(self, fh):
        """
        Create strict DBF-reader
        
        @param fh: filehandler (should be opened for binary reads)
        """
        super(YDbfStrictReader, self).__init__(fh)
        self.checkConsistency()

    def checkConsistency(self):
        """
        Some checks of DBF structure
        
        @raise AssertionError: if some check failed
        """
        ## check records
        assert self.recsize >1, "Length of record must be >1"
        if self.sig in (0x03, 0x04): 
            assert self.recsize < 4000, "Length of record must be <4000 B for dBASE III and IV"
        assert self.recsize < 32*1024, "Length of record must be <32KB"
        assert self.numrec >= 0, "Number of records must be non-negative"

        ## check fields
        assert self.numfields > 0, "The dbf file must have at least one field"
        if self.sig == 0x03:
            assert self.numfields < 128, "Number of fields in dBASE III must be <128"
        if self.sig == 0x04:
            assert self.numfields < 256, "Number of fields in dBASE IV must be <256"

        ## check fields, round 2
        for f_name, f_type, f_size, f_decimal in self.fields:
            if f_type == 'N':
                assert f_size < 20, "Size of numeral field must be <20 (field '%s', size %d)" % (f_name, f_size)
            if f_type == 'C':
                assert f_size < 255, "Size of numeral field must be <255 (field '%s', size %d)" % (f_name, f_size)
            if f_type == 'L':
                assert f_size == 1, "Size of logical field must be 1 (field '%s', size %d)" % (f_name, f_size)

        ## check size, if available
        file_name = getattr(self.fh, 'name', None)
        if file_name is not None:
            import os
            try:
                os_size = os.stat(file_name)[6]
            except OSError, msg:
                return
            dbf_size = long(self.lenheader + 1 + self.numrec*self.recsize)
            assert os_size == dbf_size

class UnicodeConverter(object):
    """
    Unicode converter which converts all strings to unicode,
    using lang code in DBF file, or implicitly specified encoding
    """
    def __init__(self, default_encoding='ascii', overwrite_encoding=False, raw_lang=0x00):
        """
        Create unicode converter for DBF reader
        
        @param default_encoding: default encoding (default ascii) for strings 
            (if lang code in DBF is not found)
        @type default_encoding: C{str}
        
        @param overwrite_encoding: overwrite lang code by default encoding, default False
        @type overwrite_encoding: C{boolean}
        
        @param raw_lang: lang code from DBF file, default 0x00 (i.e. absence of lang code)
        @type raw_lang: C{int}
        """
        if overwrite_encoding:
            self.encoding = default_encoding
        else:
            self.encoding = lib.ENCODINGS.get(raw_lang, default_encoding)

    def __call__(self, records_iterator):
        """
        Convert strings to unicode
        """
        provide_unicode = lambda x, enc: (isinstance(x, str) and unicode(x, enc)) or x
        encoding = self.encoding
        for record in records_iterator:
            yield [provide_unicode(x, encoding) for x in record]

class DictConverter(object):
    """
    Convert from list to dict each record.
    Must be the last converter in chain.
    
    @param fields_struct: structure of DBF file
    """
    def __init__(self, fields_struct):
        """
        Create dict converter
        """
        self.fields = fields_struct

    def __call__(self, records_iterator):
        """
        Convert to dict each record
        
        @param records_iterator: iterator over DBF records
        """
        fields = self.fields
        for record in records_iterator:
            record_dict = {}
            for (f_name, f_type, f_size, f_decimal), value in itertools.izip(fields, record):
                record_dict[f_name] = value
            yield record_dict
            
class YDbfReader(object):
    """
    Most common DBF reader
        @param fh: filehandler (should be opened for binary reads)
        
        @param use_unicode: use unicode instead of strings, default False
        @type use_unicode: C{boolean}
        
        @param default_encoding: default encoding for DBF file,
            default 'ascii',
            uses with C{use_unicode} option
        @type default_encoding: C{str}
        
        @param overwrite_encoding: overwrite encoding specified
            in DBF file (lang code) by default value 
            (see C{default_encoding} option), default False,
            uses with C{use_unicode} option
        @type overwrite_encoding: C{boolean}
        
        @param as_dict: represent each record as dict instead of list,
            defult False
        @type as_dict: C{boolean}
        
        @param strict: make some checks of internal DBF structure,
            default True
        @type strict: C{boolean}
    """
    
    def __init__(self, fh, **kwargs):
        """
        Create DBF reader
        """
        use_unicode = kwargs.get('use_unicode', False)
        default_encoding = kwargs.get('default_encoding', 'ascii')
        overwrite_encoding = kwargs.get('overwrite_encoding', False)
        as_dict = kwargs.get('as_dict', False)
        strict = kwargs.get('strict', True)

        if strict:
            self.reader = YDbfStrictReader(fh)
        else:
            self.reader = YDbfBasicReader(fh)

        null_converter = lambda xiter: (x for x in xiter)
        
        if use_unicode:
            self.unicode_converter = UnicodeConverter(default_encoding, overwrite_encoding, self.reader.raw_lang)
        else:
            self.unicode_converter = null_converter

        if as_dict:
            self.dict_converter = DictConverter(self.reader.fields)
        else:
            self.dict_converter = null_converter

    def __call__(self, *args, **kwargs):
        """
        Get iterator over DBF records
        """
        return self.dict_converter(self.unicode_converter(self.reader(*args, **kwargs)))

    def __len__(self): 
        return len(self.reader)

    def __getattr__(self, attr):
        if attr in self.__dict__ or attr in ('reader', 'unicode_converter', 'dict_converter',):
            return self.__dict__[attr]
        else:
            return getattr(self.__dict__['reader'], attr)
    
    def __setattr__(self, attr, val):
        if attr in self.__dict__ or attr in ('reader', 'unicode_converter', 'dict_converter',):
            self.__dict__[attr] = val
        else:
            setattr(self.__dict__['reader'], attr, val)




# -*- coding: utf-8 -*-
# YDbf - Yielded dbf reader-writer
# Inspired by code of Raymond Hettinger
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/362715
#
# Copyright (C) 2006 Yury Yurevich, Alexandr Zamaraev
# Copyright (C) 2007-2008 Yury Yurevich
#
# http://www.pyobject.ru/projects/ydbf/
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
YDbf reader-writer
"""

from ydbf.reader import YDbfReader
from ydbf.writer import YDbfWriter



# -*- coding: utf-8 -*-
# YDbf - Yielded dbf reader-writer
# Inspired by code of Raymond Hettinger
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/362715
#
# Copyright (C) 2006 Yury Yurevich, Alexandr Zamaraev
# Copyright (C) 2007-2008 Yury Yurevich
#
# http://www.pyobject.ru/projects/ydbf/
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
Common lib for both reader and writer
"""

import datetime


# Exceptions

class DbfError(Exception):
    pass

class DbfTypeError(DbfError):
    pass

# Reference data

ENCODINGS = {
    # got from dbf description [dbfspec]
    # id      name      description
    0x00:    ('ascii', 'ASCII'), # internal use
    0x01:    ('cp437', 'DOS USA'),
    0x02:    ('cp850', 'DOS Multilingual'),
    0x03:    ('cp1252', 'Windows ANSI'),
    0x04:    ('mac_roman', 'Standard Macintosh'), # NOT SHURE
    0x64:    ('cp852', 'EE MS-DOS'),
    0x65:    ('cp866', 'Russian MS-DOS'),
    0x66:    ('cp865', 'Nordic MS-DOS'),
    0x67:    ('cp861', 'Icelandic MS-DOS'),
    0x6A:    ('cp737', 'Greek MS-DOS (437G)'),
    0x6B:    ('cp857', 'Turkish MS-DOS'),
    0x96:    ('mac_cyrillic', 'Russian Macintosh'),
    0x97:    ('mac_latin2', 'Eastern Europe Macintosh'), # NOT SHURE
    0x98:    ('mac_greek', 'Greek Macinstosh'),
    0xC8:    ('cp1250', 'Windows EE'),
    0xC9:    ('cp1251', 'Russian Windows'),
    0xCA:    ('cp1254', 'Turkish Windows'),
    0xCB:    ('cp1253', 'Greek Windows'),
    ## these encodings doesn't supported by python
    # 0x68:    ('cp895', 'Kamenicky (Czech) MS-DOS'),
    # 0x69:    ('cp790', 'Mazovia (Polish) MS-DOS'),
}

REVERSE_ENCODINGS = dict([(value[0], (code, value[1])) for code, value in ENCODINGS.items()])

SIGNATURES = {
    0x02: 'FoxBase',
    0x03: 'dBASE III',
    0x04: 'dBASE IV',
    0x05: 'dBASE IV',
    0x30: 'Visual FoxPro',
    0x31: 'Visual FoxPro with AutoIncrement field',
    0x43: 'dBASE IV with SQL table and memo file',
    0x7B: 'dBASE IV with memo file',
    0x83: 'dBASE III with memo file',
    0x8B: 'dBASE IV with memo file',
    0x8E: 'dBASE IV with SQL table',
    0xB3: '.dbv and .dbt memo (Flagship)',
    0xCB: 'dBASE IV with SQL table and memo file',
    0xE5: 'Clipper SIX driver with SMT memo field',
    0xF5: 'FoxPro with memo field',
    0xFB: 'FoxPro',
}

SUPPORTED_SIGNATURES = (0x03, 0x04, 0x05)

# <   -- little endian
# B   -- version number (signature)
# 3B  -- last update (YY, MM, DD)
# L   -- number of records
# H   -- length of header
# H   -- length of each record
# 17x -- pad (2B -- reserved, 
#              B -- incomplete transaction, 
#              B -- encryption flag, 
#             4B -- free record thread (reserved for LAN)
#             8B -- reserved for multiuser dBASE
#              B -- MDX flag)
# B   -- language driver
# 2x  -- pad (2B -- reserved)
HEADER_FORMAT = '<B3BLHH17xB2x'

# <   -- little endian
# 11s -- field name in ASCII (terminated by 0x00)
# c   -- field type (ASCII)
# 4x  -- field data address ( 2B -- address in memory (for dBASE) 
#                          OR 4B -- offset of field from 
#                                   beginning of record (for FoxPro)
# B   -- field length
# B   -- decimal count
# 14x -- pad (2B -- reserved for multi-user dBASE,
#              B -- work area id ()
#             2B -- reserved for multi-user dBASE,
#              B -- flag for SET FIELDS
#             7B -- reserved
#              B -- index field flag)
# B   -- language driver
# 2x  -- pad (2B -- reserved)
FIELD_DESCRIPTION_FORMAT = '<11sc4xBB14x'

# Common functions

def dbf2date(dbf_str):
    '''
    Convert date from dbf-string to datetime.date
    @param dbf_str: string in format YYYYMMDD
    @return: datetime.date instance
    '''
    if dbf_str is None or not dbf_str.isdigit() or len(dbf_str) != 8:
        result = None
    else:
        result = datetime.date(int(dbf_str[:4]), int(dbf_str[4:6]), int(dbf_str[6:8]))
    return result

def date2dbf(dt):
    '''
    Convert date from datetime.date to dbf-string
    @param dt: datetime.date instance
    @return: string in format YYYYMMDD
    '''
    if not isinstance(dt, datetime.date):
        raise TypeError("Espects datetime.date instead of %s" % type(dt))
    return "%04d%02d%02d" % (dt.year, dt.month, dt.day)
    
def dbf2str(dbf_str):
    '''
    Convert date from dbf-string to string (DD.MM.YYYY)
    @param dbf_str: string in format YYYYMMDD
    @return: string in format DD.MM.YYYY
    '''
    if dbf_str is None or not dbf_str.isdigit() or len(dbf_str) != 8:
        result = None
    else:
        result = ".".join( reversed( (dbf_str[:4], dbf_str[4:6], dbf_str[6:8]) ) )
    return result

def str2dbf(dt_str):
    '''
    Convert from string to dbf-string date
    @param dt_str: string in format DD.MM.YYYY
    @return: dbf-string date (string in format YYYYMMDD)
    '''
    if not isinstance(dt_str, basestring):
        raise TypeError("Espects string or unicode instead of %s" % type(dt_str))
    str_l = len(dt_str)
    if str_l != 10:
        raise ValueError('Datestring must be 10 symbols (DD.MM.YYYY) length instead of %d' % str_l)    
    d, m, y = dt_str.split('.')
    return ''.join((y,m,d))

# References:
# [dbfspec]: http://www.clicketyclick.dk/databases/xbase/format/index.html




# -*- coding: utf-8 -*-
# YDbf - Yielded dbf reader-writer
# Inspired by code of Raymond Hettinger
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/362715
#
# Copyright (C) 2006 Yury Yurevich, Alexandr Zamaraev
# Copyright (C) 2007-2008 Yury Yurevich
#
# http://www.pyobject.ru/projects/ydbf/
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
"""
DBF writer
"""
__all__ = ["YDbfBasicWriter", "YDbfWriter"]

import struct
import itertools
import datetime

from ydbf import lib

class YDbfBasicWriter(object):
    """
    Class for writing DBF.
    
        @param fh: filehandler, should be opened for binary writing
        @param fields: fields structure in format
            [(NAME, TYP, SIZE, DEC), ...]
        @type NAME: string
        @type TYP: string in ("N", "D", "C", "L")
        @type SIZE: integer
        @type DEC: integer
    """
    def __init__(self, fh, fields, as_dict=False):
        """
        Constructor
        """
        self.now = datetime.date.today()
        self.numrec = 0
        self.numfields = len(fields)
        self.lenheader = self.numfields * 32 + 33
        self.recsize = sum([field[2] for field in fields]) + 1
        self.date2dbf = lib.date2dbf
        self.sig = 0x03  # signature, DBF 3
        self.lang = 0x0 # w/o lang code
        self.fh = fh
        self.fields = fields
            
    def writeHeader(self):
        """
        Write DBF-header
        """
        pos = self.fh.tell()
        self.fh.seek(0)
        year, month, day = self.now.year-1900, self.now.month, self.now.day
        self.hdr = struct.pack(lib.HEADER_FORMAT, self.sig, year, month,
                               day, self.numrec, self.lenheader,
                               self.recsize, self.lang)
        self.fh.write(self.hdr)
        for name, typ, size, deci in self.fields:
            name = name.ljust(11).replace(' ', '\x00')   # compability with py23, for py24 it looks like name.ljust(11, '\x00')
            fld = struct.pack(lib.FIELD_DESCRIPTION_FORMAT, name, typ, size, deci)
            self.fh.write(fld)
        # terminator
        self.fh.write('\r')
        if pos > 0:
            self.fh.seek(pos)
    
    def __call__(self, records):
        """
        Run DBF-creator
        @param records: iterator over records 
            (each record is tuple or sequense of values)
        """
        self.writeHeader()
        i = 0
        for record in records:
            i += 1
            self.fh.write(' ')                        # deletion flag
                
            for (name, typ, size, deci), value in itertools.izip(self.fields, record):
                if typ ==   'N':
                    if value is None:
                        value = 0
                    if deci:
                        pattern = "%%.%df" % deci # ->%.2f 
                        value = pattern % value
                    else:
                        value = int(value)
                    value = str(value).rjust(size)
                elif typ == 'D':
                    if value is None:
                        value = ' '*8
                    else:
                        value = self.date2dbf(value)
                elif typ == 'L':
                    value = str(value)[0].upper()
                else:
                    if value is None:
                        value = ''
                    value = str(value)[:size].ljust(size)
                if len(value) != size:
                    raise ValueError("value '%s' of %s not aligned to %d" % (str(value), name, size))
                self.fh.write(value)
        self.numrec = i
        self.writeHeader()
        # End of file
        self.fh.write('\x1A')
        self.fh.flush()

class DictDeconverter(object):
    """
    Deconvert from dict to list
    
    Must be the last deconverter in chain
    
    @param fields_struct: structure of DBF file
    """
    
    def __init__(self, fields_struct):
        """
        Create dict deconverter
        """
        self.fields = fields_struct
    
    def __call__(self, records_iterator):
        """
        Deconvert from dict each record
        
        @param records_iterator: iterator over DBF records
        """
        for record in records_iterator:
            yield [record[f_name] for f_name, f_type, f_size, f_decimal in self.fields]

class UnicodeDeconverter(object):
    """
    Deconverts all unicode-strings to byte-strings,
    using lang code in DBF file, or implicitly specified encoding
    """
    def __init__(self, encoding='ascii'):
        """
        Create unicode deconverter for DBF writer
        
        @param encoding: encoding (default ascii, 0x0 lang code in DBF) for strings 
        @type encoding: C{str}
        """
        self.encoding = encoding

    def __call__(self, records_iterator):
        """
        Deconvert strings from unicode
        """
        provide_str = lambda x, enc: (isinstance(x, unicode) and x.encode(encoding)) or x
        encoding = self.encoding
        for record in records_iterator:
            yield [provide_str(x, encoding) for x in record]

class YDbfWriter(object):
    """
    Most common DBF writer
        @param fh: filehandler (should be opened for binary reads)
        
        @param use_unicode: use unicode instead of strings, default False
        @type use_unicode: C{boolean}
        
        @param encoding: encoding for DBF file,
            default 'ascii' (or 0x0 war lang code),
        @type encoding: C{str}
        
        @param raw_lang: raw lang code in DBF file
            by default it taken from encoding, but
            using this option you can override it
        @type C{int}
        
        @param as_dict: represent each record as dict instead of list,
            defult False
        @type as_dict: C{boolean}
        
    """

    def __init__(self, fh, fields, **kwargs):
        """
        Create DBF writer
        """
        use_unicode = kwargs.get('use_unicode', False)
        encoding = kwargs.get('encoding', 'ascii')
        raw_lang = kwargs.get('raw_lang')
        if raw_lang is None:
            raw_lang = lib.REVERSE_ENCODINGS.get(encoding, (0x0, 'ASCII'))[0]
        as_dict = kwargs.get('as_dict', False)
        
        self.writer = YDbfBasicWriter(fh, fields)
        self.writer.lang = raw_lang
        
        null_deconverter = lambda xiter: (x for x in xiter)
        
        if use_unicode:
            self.unicode_deconverter = UnicodeDeconverter(encoding)
        else:
            self.unicode_deconverter = null_deconverter
        
        if as_dict:
            self.dict_deconverter = DictDeconverter(fields)
        else:
            self.dict_deconverter = null_deconverter
    
    def __call__(self, records, *args, **kwargs):
        """
        Write records to DBF
        """
        self.writer(self.unicode_deconverter(self.dict_deconverter(records, *args, **kwargs)))
    
    def __getattr__(self, attr):
        if attr in self.__dict__ or attr in ('writer', 'unicode_deconverter', 'dict_deconverter',):
            return self.__dict__[attr]
        else:
            return getattr(self.__dict__['writer'], attr)
    
    def __setattr__(self, attr, val):
        if attr in self.__dict__ or attr in ('writer', 'unicode_deconverter', 'dict_deconverter',):
            self.__dict__[attr] = val
        else:
            setattr(self.__dict__['writer'], attr, val)

    
    # backwards compability
    def get_now(self): return self.writer.now
    def set_now(self, value): self.writer.now = value
    now = property(get_now, set_now)
        



from setuptools import setup, find_packages
import sys, os

version = '0.0.1'

setup(name='YDbf',
      version=version,
      description="Yielded Dbf reader and writer",
      long_description="""\
Pythonic Dbf reader and writer""",
      classifiers=[
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: GNU General Public License (GPL)',
      'Operating System :: OS Independent',
      'Programming Language :: Python',
      'Topic :: Database',
      'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords='',
      author='Yury Yurevich',
      author_email='the.pythy@gmail.com',
      url='http://www.pyobject.ru/projects/YDbf',
      license='GNU GPL2',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      install_requires=[],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )
      
